import torch 
import torch.nn as nn
import torch.nn.functional as F
import textwrap
import matplotlib.pyplot as plt  
from .esm2 import ESM2 
from .create_dataset import esm_alphabet, convert, idx2token  
import torch

esm_model = ESM2()  
device = 'cuda' if torch.cuda.is_available() else 'cpu'  


class AffinityPredictor(nn.Module):
    def __init__(self):
        super().__init__()
        self.encoder = esm_model
        self.predictor = nn.Sequential(
            nn.Linear(1280,1)
        )
        
        
    def forward(self, seq, output_type='affinity'):
        res = self.encoder(seq, repr_layers=[33], return_contacts=False)
        rep = res['representations'][33]
        rep = rep.mean(1)
        y = self.predictor(rep) 

        if output_type == 'affinity':
            return y
        elif output_type == 'likelihood':
            return F.softmax(res['logits'], dim=-1)
    
    
    def make_scFv(self, heavy_seq, light_seq):
        return heavy_seq + 'GGGGSGGGGSGGGGS' + light_seq 
    
    
    def get_affinity(self, sequences, batch_size=4):
        if isinstance(sequences, str):
            sequences = [sequences]
        with torch.no_grad():
            seqs = [torch.tensor(convert(seq)) for seq in sequences]
            seqs_tensor = torch.stack(seqs).to(device)
            num_batches = (len(sequences) + batch_size - 1) // batch_size
            predicted_affinities = []
            for i in range(num_batches):
                start_idx = i * batch_size
                end_idx = min((i + 1) * batch_size, len(sequences))
                batch_tensor = seqs_tensor[start_idx:end_idx]
                outputs_batch = self(batch_tensor)
                batch_affinities = outputs_batch 
                predicted_affinities.extend(batch_affinities)
        if len(predicted_affinities) == 1:
            return predicted_affinities[0]
        else:
            predicted_affinities = torch.cat(predicted_affinities, dim=0)
            return predicted_affinities 
    
    
    def get_embeddings(self, sequences, mode='res', batch_size=4):
        if mode not in ['res', 'seq']:
            print('Wrong mode selected. Available modes are \'res\' or \'seq\'.')
            return 
        if isinstance(sequences, str):
            sequences = [sequences]
        seqs = [torch.tensor(convert(seq)) for seq in sequences]
        with torch.no_grad():
            seqs_tensor = torch.stack(seqs).to(device)
            num_batches = (len(sequences) + batch_size - 1) // batch_size
            embeddings_batch = []
            for i in range(num_batches):
                start_idx = i * batch_size
                end_idx = min((i + 1) * batch_size, len(sequences))
                batch_tensor = seqs_tensor[start_idx:end_idx]
                outputs_batch = self.encoder(batch_tensor, repr_layers=[33], return_contacts=False)
                batch_embeddings = outputs_batch['representations'][33]
                if mode == 'seq':
                    batch_embeddings = batch_embeddings.mean(dim=1)
                embeddings_batch.extend(batch_embeddings)
        if len(embeddings_batch) == 1:
            return embeddings_batch[0]
        else:
            return embeddings_batch 
    
    
    def get_contact_map(self, input_sequence, mode='VH-VL', plot=True):
        def wrap_text(text, width):
            return "\n".join(textwrap.wrap(text, width))
        if mode not in ['VH-VL', 'scFv', None]: 
            print('Wrong mode selected. Available modes are \'VH-VL\' or \'scFv\'.')
            return 
        seq_txt = wrap_text(input_sequence, 85)  
        with torch.no_grad():
            toks = convert(input_sequence)
            seq = torch.unsqueeze(torch.tensor(toks),0).to(device) 
            outputs_batch =  self.encoder(seq, repr_layers=[33], return_contacts=True) 

        AAs = [idx2token(idx) for idx in toks][1:-1]
        contacts = outputs_batch['contacts'].to('cpu').numpy()
        combined_contacts = contacts[0][1:-1, 1:-1] 

        HC_last_idx = input_sequence.find('GGGGSGGGGSGGGGS')
        LC_length = len(input_sequence)-(HC_last_idx)-15
        LC_first_idx = HC_last_idx + 15 
        LC_last_idx = LC_first_idx+LC_length 
        
        if mode == 'scFv':
            last_idx = AAs.index('<pad>') 
            plt.subplots(1,1, figsize=(30, 30)) 
            contacts = combined_contacts[:last_idx, :last_idx]
            labels = AAs[:last_idx]  
            x_labels = [labels[i]+str(i+1) for i in range(len(labels))]
            im = plt.imshow(contacts, cmap='Blues', aspect='auto', extent=[-0.5, len(labels) - 0.5, -0.5, len(labels) - 0.5], origin="lower")
            plt.xticks(range(len(labels)), x_labels, rotation=90, ha="center", fontsize=10)
            plt.yticks(range(len(labels)), x_labels)
            plt.title(f'Contact Maps for input antibody sequence:\n {seq_txt}', fontsize=18) 
            plt.tight_layout() 
            plt.show() 
            contactss = combined_contacts[:last_idx, :last_idx] 
            
            
        elif mode == 'VH-VL':
            fig, ax = plt.subplots(1,2, figsize=(40, 20))  
            left_top_contacts = combined_contacts[:HC_last_idx, :HC_last_idx]
            labels = AAs[:HC_last_idx] 
            x_labels = [labels[i]+str(i+1) for i in range(len(labels))]
            im = ax[0].imshow(left_top_contacts, cmap='Blues', aspect='auto', extent=[-0.5, len(labels) - 0.5, -0.5, len(labels) - 0.5], origin="lower")
            ax[0].set_xticks(range(HC_last_idx))
            ax[0].set_yticks(range(HC_last_idx))
            ax[0].set_xticklabels(x_labels, rotation=90, ha="center", fontsize=10)   
            ax[0].set_yticklabels(x_labels, fontsize=10)
            ax[0].set_title('Heavy Chain contacts', fontsize=18)
        
            LC_contacts = combined_contacts[LC_first_idx:LC_last_idx, LC_first_idx:LC_last_idx]
            labels = AAs[LC_first_idx:LC_last_idx]  
            x_labels = [labels[i]+str(i+1) for i in range(len(labels))]
            im = ax[1].imshow(LC_contacts, cmap='Blues', aspect='auto', extent=[-0.5, len(labels) - 0.5, -0.5, len(labels) - 0.5], origin="lower")
            ax[1].set_xticks(range(LC_length))
            ax[1].set_yticks(range(LC_length))
            ax[1].set_xticklabels(x_labels, rotation=90, ha="center", fontsize=10)  
            ax[1].set_yticklabels(x_labels, fontsize=10)
            ax[1].set_title('Ligth Chain contacts', fontsize=18)
            plt.suptitle(f'Contact Maps for input antibody sequence:\n {seq_txt}', fontsize=18)
            plt.tight_layout() 
            plt.show() 
            contactss = combined_contacts[:LC_last_idx, :LC_last_idx] 
        
        elif mode == None:
            last_idx = AAs.index('<pad>') 
            contactss = combined_contacts[:LC_last_idx, :LC_last_idx] 
        return contactss 
   


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data.sampler import SubsetRandomSampler
import numpy as np 

# Create dataset from sequences 
esm_alphabet = ['<cls>', '<pad>', '<eos>', '<unk>', 'L',  'A', 'G', 'V', 'S', 'E', 'R', 'T', 'I', 'D',
 'P', 'K', 'Q', 'N', 'F', 'Y', 'M', 'H', 'W', 'C', 'X', 'B', 'U', 'Z', 'O', '.', '-', '<null_1>', '<mask>']

token2idx_dict = dict(zip(esm_alphabet, list(range(len(esm_alphabet)))))

def token2idx(token):
    if token in token2idx_dict:
        token = token2idx_dict[token]
    else:
        token = token2idx_dict['<unk>']
    return token

def idx2token(idx):
    for token, token_idx in token2idx_dict.items():
        if idx == token_idx:
            return token
    
def convert(seq, length=256):
    tokens = [token2idx('<cls>')] + [1] * length + [token2idx('<eos>')]
    if len(seq) > length:
        start = np.random.randint(len(seq)-length)
        seq = seq[start: start+length]
    for i, tok in enumerate(seq):
        tokens[i+1] = token2idx(tok)
    return np.array(tokens, dtype=int)


class SeqDataset(torch.utils.data.Dataset):
    def __init__(self, ab_seq, pred_affinity) -> None:
        super().__init__()
        self.ab_seq = ab_seq
        self.pred_affinity = pred_affinity 
        
    def __getitem__(self, idx):
        selected_ab = self.ab_seq[idx]
        affinity_val = self.pred_affinity[idx]
        return convert(selected_ab), float(affinity_val) 
    
    def __len__(self):
        return len(self.ab_seq) 


def get_dataloader(dataset, batch_size=8, num_workers=1):
    data_loader = torch.utils.data.DataLoader(
        dataset, 
        batch_size=batch_size, 
        shuffle=True,  # Set to True if you want to shuffle the data
        num_workers=num_workers,
        pin_memory=True  # Set to True if you have a GPU
    )

    return data_loader



def get_dataloaders(dataset, batch_size=8, seed=42, train_frac=0.8, num_workers=1):
    random_seed = seed 
    batch_size = batch_size 
    dataset_size = len(dataset)
    train_sample_num = int(dataset_size * train_frac)

    indices = list(range(dataset_size))

    np.random.seed(random_seed)
    np.random.shuffle(indices)

    train_indices = indices[:train_sample_num]
    test_indices = indices[train_sample_num:]

    train_sampler = SubsetRandomSampler(train_indices)
    test_sampler = SubsetRandomSampler(test_indices)

    train_loader = torch.utils.data.DataLoader(
        dataset, 
        batch_size=batch_size, 
        sampler=train_sampler, 
        num_workers=num_workers,
        pin_memory=False)

    test_loader = torch.utils.data.DataLoader(
        dataset, 
        batch_size=batch_size, 
        sampler=test_sampler, 
        num_workers=num_workers,
        pin_memory=False)

    return train_loader, test_loader 




from setuptools import setup, find_packages 

setup(
    name='AbAffinity',
    version='0.1.2', 
    description='Predict Affinity against SARS-CoV-2 HR2 peptide',
    author='Faisal Bin Ashraf',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown', 
    packages=find_packages(),
    include_package_data=True, 
    package_data={
        '': ['ESM2/*.py'],             # Include all .py files in EM2 folder
    },
    install_requires=[
        'torch',              # PyTorch
        'numpy',              # For numerical operations
        'matplotlib',         # For plotting graphs
        'textwrap3',           # For wrapping text
    ],
    python_requires='>=3.6',
)



import torch  
import numpy as np 
import matplotlib.pyplot as plt  
from .create_dataset import esm_alphabet, convert  
from .affinity_pred_model import AffinityPredictor
import textwrap
import random 
from collections import OrderedDict


device = 'cuda' if torch.cuda.is_available() else 'cpu'  

def AbAffinity():
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    model = AffinityPredictor().to(device) 
    state_dict = torch.load('model/2024-01-20_best_mse.pth',  map_location=device)
    # Remove the 'module.' prefix from keys
    new_state_dict = OrderedDict()
    for k, v in state_dict.items():
        if k.startswith('module.'):
            new_state_dict[k[7:]] = v  # Remove 'module.'
        else:
            new_state_dict[k] = v
    model.load_state_dict(new_state_dict)
    model = model.eval() 
    return model 

